import os
import warnings
from typing import IO, Dict, Optional, Callable
from abc import ABC, abstractmethod

from ._utils import atomic_lambda_save
from .callbacks import Callback


class PeriodicSaveCallback(ABC, Callback):
    """
    Write a file (or checkpoint) after every epoch. ``filename`` can contain named formatting options,
    which will be filled the value of ``epoch`` and keys in ``logs`` (passed in ``on_epoch_end``). For
    example: if ``filename`` is ``weights.{epoch:02d}-{val_loss:.2f}.txt``, then ``save_file()`` will be
    called with a file descriptor for a file with the epoch number and the validation loss in the
    filename.

    By default, the file is written atomically to the specified filename so that the training can
    be killed and restarted later using the same filename for periodic file saving. To do so, a
    temporary file is created with the name of ``filename + '.tmp'`` and is then moved to the final
    destination after the checkpoint is done. The ``temporary_filename`` argument allows to change the
    path of this temporary file.

    Args:
        filename (str): Path to save the model file.
        monitor (str): Quantity to monitor.
            (Default value = 'val_loss')
        verbose (bool): Whether to display a message when saving and restoring a checkpoint.
            (Default value = False)
        save_best_only (bool): If `save_best_only` is true, the latest best model according to the
            quantity monitored will not be overwritten.
            (Default value = False)
        keep_only_last_best (bool): Whether only the last saved best checkpoint is kept. Applies only when
             `save_best_only` is true.
             (Default value = False)
        restore_best (bool): If `restore_best` is true, the model will be reset to the last best checkpoint done.
            This option only works when `save_best_only` is also true. (Default value = False)
        mode (str): One of {'min', 'max'}.
            If `save_best_only` is true, the decision to overwrite the current save file is made based
            on either the maximization or the minimization of the monitored quantity. For
            `val_accuracy`, this should be `max`, for `val_loss` this should be `min`, etc.
            (Default value = 'min')
        period (int): Interval (number of epochs) between checkpoints.
            (Default value = 1)
        temporary_filename (str, optional): Temporary filename for the checkpoint so that the
            last checkpoint can be written atomically. See the ``atomic_write`` argument.
        atomic_write (bool): Whether to write atomically the checkpoint. See the description above
            for details.
            (Default value = True)
        open_mode (str): ``mode`` option passed to :func:`open()`.
            (Default value = 'wb')
    """

    def __init__(
        self,
        filename: str,
        *,
        monitor: str = 'val_loss',
        mode: str = 'min',
        save_best_only: bool = False,
        keep_only_last_best: bool = False,
        restore_best: bool = False,
        period: int = 1,
        verbose: bool = False,
        temporary_filename: Optional[str] = None,
        atomic_write: bool = True,
        open_mode: str = 'wb',
        read_mode: str = 'rb',
    ) -> None:
        super().__init__()
        self.filename = filename
        self.monitor = monitor
        self.verbose = verbose
        self.save_best_only = save_best_only
        self.keep_only_last_best = keep_only_last_best
        self.restore_best = restore_best
        self.temporary_filename = temporary_filename
        self.atomic_write = atomic_write
        self.open_mode = open_mode
        self.read_mode = read_mode
        self.best_filename = None

        if self.keep_only_last_best and not self.save_best_only:
            raise ValueError("The 'keep_only_last_best' argument only works when 'save_best_only' is also true.")
        if self.restore_best and not self.save_best_only:
            raise ValueError("The 'restore_best' argument only works when 'save_best_only' is also true.")

        if self.save_best_only:
            if mode not in ['min', 'max']:
                raise ValueError(f"Invalid mode '{mode}'")
            if mode == 'min':
                self.monitor_op = lambda x, y: x < y
                self.current_best = float('Inf')
            elif mode == 'max':
                self.monitor_op = lambda x, y: x > y
                self.current_best = -float('Inf')

        self.period = period

    @abstractmethod
    def save_file(self, fd: IO, epoch_number: int, logs: Dict) -> None:
        """
        Abstract method that is called every time a save needs to be done.

        Args:
            fd (IO): The descriptor of the file in which to write.
            epoch_number (int): The epoch number.
            logs (Dict): Dictionary passed on epoch end.
        """
        pass

    def _save_file(self, filename: str, epoch_number: int, logs: Dict) -> None:
        atomic_lambda_save(
            filename,
            self.save_file,
            (epoch_number, logs),
            temporary_filename=self.temporary_filename,
            open_mode=self.open_mode,
            atomic=self.atomic_write,
        )

    def on_epoch_end(self, epoch_number: int, logs: Dict) -> None:
        filename = self.filename.format_map(logs)

        if self.save_best_only:
            if self.monitor in logs:
                if self.monitor_op(logs[self.monitor], self.current_best):
                    old_best = self.current_best
                    self.current_best = logs[self.monitor]
                    old_best_filename = self.best_filename
                    self.best_filename = filename

                    if self.verbose:
                        print(
                            f'Epoch {epoch_number:d}: {self.monitor} improved from {old_best:0.5f} '
                            f'to {self.current_best:0.5f}, saving file to {self.best_filename}'
                        )
                    self._save_file(self.best_filename, epoch_number, logs)
                    if (
                        self.keep_only_last_best
                        and self.best_filename != old_best_filename
                        and old_best_filename is not None
                    ):
                        os.remove(old_best_filename)
            else:
                raise KeyError(f"The monitored metric name {self.monitor} is not found in computed metrics.")
        elif epoch_number % self.period == 0:
            if self.verbose:
                print(f'Epoch {epoch_number:d}: saving file to {filename}')
            self._save_file(filename, epoch_number, logs)

    @abstractmethod
    def restore(self, fd: IO) -> None:
        """
        Abstract method that is called when a save needs to be restored. This happens at the end of the training when
        ``restore_best`` is true.

        Args:
            fd (IO): The descriptor of the file to read.
        """
        pass

    def on_train_end(self, logs: Dict) -> None:
        if self.restore_best:
            if self.best_filename is not None:
                if self.verbose:
                    print(f'Restoring data from {self.best_filename}')
                # pylint: disable=unspecified-encoding
                open_kwargs = dict(encoding='utf-8') if 'b' not in self.read_mode else {}
                with open(self.best_filename, self.read_mode, **open_kwargs) as fd:
                    self.restore(fd)
            else:
                warnings.warn('No data to restore!')


class PeriodicSaveLambda(PeriodicSaveCallback):
    """
    Call a lambda with a file descriptor after every epoch. See
    :class:`~ enaptorch.PeriodicSaveCallback` for the arguments' descriptions.

    Args:
        func (Callable[[fd, int, dict], None]): The lambda that will be called with a file descriptor, the
            epoch number and the epoch logs.
        restore (Callable[[fd], None]): The lambda that will be called with a file descriptor to restore
            the state if necessary.

    See:
        :class:`~ enaptorch.PeriodicSaveCallback`
    """

    def __init__(self, func: Callable, *args, restore: Optional[Callable] = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.func = func
        self._restore = restore

    def save_file(self, fd: IO, epoch_number: int, logs: Dict) -> None:
        self.func(fd, epoch_number, logs)

    def restore(self, fd: IO) -> None:
        if self._restore is not None:
            self._restore(fd)



import torch.nn.functional as F

from .metrics_registering import do_register_metric_func

do_register_metric_func(F.cross_entropy)
do_register_metric_func(F.kl_div)
do_register_metric_func(F.binary_cross_entropy, ['binary_cross_entropy', 'bce'], unique_name='bce')
do_register_metric_func(
    F.binary_cross_entropy_with_logits,
    ['binary_cross_entropy_with_logits', 'bce_with_logits'],
    unique_name='bce_with_logits',
)
do_register_metric_func(F.poisson_nll_loss, unique_name='poisson_nll')
do_register_metric_func(F.hinge_embedding_loss, unique_name='hinge_embedding')
do_register_metric_func(F.l1_loss, unique_name='l1')
do_register_metric_func(F.mse_loss, unique_name='mse')
do_register_metric_func(F.multilabel_margin_loss, unique_name='multilabel_margin')
do_register_metric_func(F.multilabel_soft_margin_loss, unique_name='multilabel_soft_margin')
do_register_metric_func(F.multi_margin_loss, unique_name='multi_margin')
do_register_metric_func(F.nll_loss, unique_name='nll')
do_register_metric_func(F.smooth_l1_loss, unique_name='smooth_l1')
do_register_metric_func(F.soft_margin_loss, unique_name='soft_margin')



 
# pylint: disable=too-many-locals
import os
from unittest.mock import MagicMock, call, ANY

import numpy as np
import torch
import torch.nn as nn

from  enaptorch import Callback
from tests.framework.base import CaptureOutputBase


class ModelFittingTestCase(CaptureOutputBase):
    epochs = 10
    steps_per_epoch = 5
    batch_size = 20

    evaluate_dataset_len = 107

    cuda_device = int(os.environ.get('CUDA_DEVICE', 0))

    def setUp(self):
        self.mock_callback = MagicMock(spec=Callback)
        self.batch_metrics = []
        self.batch_metrics_names = []
        self.batch_metrics_values = []
        self.epoch_metrics = []
        self.epoch_metrics_names = []
        self.epoch_metrics_values = []
        self.model = None
        self.pytorch_network = None
        self.optimizer = None

    def _get_callback_expected_on_calls_when_training(self, params, logs, has_valid=True, steps=None, valid_steps=10):
        # pylint: disable=too-many-arguments
        if steps is None:
            steps = params['steps']

        train_batch_dict = dict(zip(self.batch_metrics_names, self.batch_metrics_values), time=ANY, loss=ANY)
        train_epochs_dict = dict(zip(self.epoch_metrics_names, self.epoch_metrics_values))
        log_dict = {**train_batch_dict, **train_epochs_dict}
        if has_valid:
            val_batch_metrics_names = ['val_' + metric_name for metric_name in self.batch_metrics_names]
            val_batch_dict = dict(zip(val_batch_metrics_names, self.batch_metrics_values), val_loss=ANY)
            val_epoch_metrics_names = ['val_' + metric_name for metric_name in self.epoch_metrics_names]
            val_epochs_dict = dict(zip(val_epoch_metrics_names, self.epoch_metrics_values))
            log_dict.update({**val_batch_dict, **val_epochs_dict})

        for epoch, log in enumerate(logs, 1):
            self.assertEqual(log, dict(log_dict, epoch=epoch))

        call_list = []
        call_list.append(call.on_train_begin({}))
        for epoch in range(1, params['epochs'] + 1):
            call_list.append(call.on_epoch_begin(epoch, {}))
            for step in range(1, steps + 1):
                call_list.append(call.on_train_batch_begin(step, {}))
                call_list.append(call.on_backward_end(step))
                call_list.append(call.on_train_batch_end(step, {'batch': step, 'size': ANY, **train_batch_dict}))
            if has_valid:
                call_list.append(call.on_valid_begin({}))
                for step in range(1, valid_steps + 1):
                    call_list.append(call.on_valid_batch_begin(step, {}))
                    call_list.append(
                        call.on_valid_batch_end(step, {'batch': step, 'size': ANY, 'time': ANY, **val_batch_dict})
                    )
                call_list.append(call.on_valid_end({'time': ANY, **val_batch_dict, **val_epochs_dict}))
            call_list.append(call.on_epoch_end(epoch, logs[epoch - 1]))

        call_list.append(call.on_train_end({}))
        return call_list

    def _test_callbacks_train(self, params, logs, *args, **kwargs):
        self.assertEqual(len(logs), params['epochs'])

        call_list = self._get_callback_expected_on_calls_when_training(params, logs, *args, **kwargs)

        method_calls = self.mock_callback.method_calls
        self.assertIn(call.set_model(self.model), method_calls[:2])  # skip set_model and set param call
        self.assertIn(call.set_params(params), method_calls[:2])

        self.assertEqual(len(method_calls), len(call_list) + 2)  # for set_model and set param
        self.assertEqual(method_calls[2:], call_list)

    def _get_callback_expected_on_calls_when_testing(self, params):
        test_batch_dict = {"time": ANY, "test_loss": ANY}
        test_batch_dict.update(
            {
                "test_" + metric_name: metric
                for metric_name, metric in zip(self.batch_metrics_names, self.batch_metrics_values)
            }
        )

        call_list = []
        call_list.append(call.on_test_begin({}))
        for batch in range(1, params['steps'] + 1):
            call_list.append(call.on_test_batch_begin(batch, {}))
            call_list.append(call.on_test_batch_end(batch, {'batch': batch, 'size': ANY, **test_batch_dict}))

        test_batch_dict.update(
            {
                "test_" + metric_name: metric
                for metric_name, metric in zip(self.epoch_metrics_names, self.epoch_metrics_values)
            }
        )
        call_list.append(call.on_test_end({"time": ANY, "test_loss": ANY, **test_batch_dict}))
        return call_list

    def _test_callbacks_test(self, params):
        call_list = self._get_callback_expected_on_calls_when_testing(params)

        method_calls = self.mock_callback.method_calls
        self.assertEqual(call.set_model(self.model), method_calls[0])  # skip set_model and set param call
        self.assertEqual(call.set_params(params), method_calls[1])

        self.assertEqual(len(method_calls), len(call_list) + 2)  # for set_model and set param
        self.assertEqual(method_calls[2:], call_list)

    def _test_return_dict_logs(self, logs):
        test_logs = {"time": ANY, "test_loss": ANY}
        test_logs.update(
            {
                "test_" + metric_name: metric
                for metric_name, metric in zip(self.batch_metrics_names, self.batch_metrics_values)
            }
        )
        test_logs.update(
            {
                "test_" + metric_name: metric
                for metric_name, metric in zip(self.epoch_metrics_names, self.epoch_metrics_values)
            }
        )
        self.assertEqual(logs, test_logs)

    def _test_size_and_type_for_generator(self, pred_y, expected_size):
        if isinstance(pred_y, (list, tuple)):
            for o in pred_y:
                self._test_size_and_type_for_generator(o, expected_size)
        elif isinstance(pred_y, dict):
            for val in pred_y.values():
                self._test_size_and_type_for_generator(val, expected_size)
        else:
            self.assertEqual(type(pred_y), np.ndarray)
            self.assertEqual(pred_y.shape, expected_size)

    def _test_device(self, device):
        for p in self.pytorch_network.parameters():
            self.assertEqual(p.device, device)

        for v in self.optimizer.state.values():
            if torch.is_tensor(v):
                self.assertEqual(v.device, device)

        for param_group in self.optimizer.param_groups:
            for param in param_group['params']:
                if torch.is_tensor(param):
                    self.assertEqual(param.device, device)


class MultiIOModel(nn.Module):
    """Model to test multiple inputs/outputs"""

    def __init__(self, num_input=2, num_output=2):
        super().__init__()
        inputs = []
        for _ in range(num_input):
            inputs.append(nn.Linear(1, 1))
        self.inputs = nn.ModuleList(inputs)

        outputs = []
        for _ in range(num_output):
            outputs.append(nn.Linear(num_input, 1))
        self.outputs = nn.ModuleList(outputs)

    def forward(self, *x):
        inp_to_cat = []
        for i, inp in enumerate(self.inputs):
            inp_to_cat.append(inp(x[i]))
        inp_cat = torch.cat(inp_to_cat, dim=1)

        outputs = []
        for out in self.outputs:
            outputs.append(out(inp_cat))

        outputs = outputs if len(outputs) > 1 else outputs[0]
        return outputs



 
# pylint: disable=protected-access
from unittest import TestCase
from unittest.mock import patch

from  enaptorch import EmptyStringAttrClass, ColorProgress


class EmptyStringAttrClassTest(TestCase):
    def test_empty_string_att_class_get_attr_return_empty_string(self):
        actual = getattr(EmptyStringAttrClass(), "an_attribute")
        expected = ""
        self.assertEqual(actual, expected)


class ColorProgressTest(TestCase):
    @patch(" enaptorch.framework.color_formatting.colorama", None)
    def test_no_colorama_and_coloring_raise_warning(self):
        with self.assertWarns(ImportWarning):
            ColorProgress(coloring=True)

    def test_format_duration_with_days(self):
        # 2 days * 24 hours * 60 minutes * 60 seconds
        a_two_days_duration = 2 * 24 * 60 * 60
        color_progress = ColorProgress(coloring=False)
        actual = color_progress._format_duration(a_two_days_duration)
        expected = "2d0h0m0.00s"

        self.assertEqual(actual, expected)

        # 35 hours (1 day and 12 hours) + 30 minutes + 30 seconds
        a_day_and_12_hours_30_minutes_30_seconds_duration = 36 * 60 * 60 + 30 * 60 + 30
        color_progress = ColorProgress(coloring=False)
        actual = color_progress._format_duration(a_day_and_12_hours_30_minutes_30_seconds_duration)
        expected = "1d12h30m30.00s"

        self.assertEqual(actual, expected)

    @patch(" enaptorch.framework.color_formatting.jupyter", True)
    def test_do_update_when_jupyter_return_true(self):
        color_progress = ColorProgress(coloring=False)

        self.assertTrue(color_progress._do_update())

    def test_do_update_when_no_coloring_return_true(self):
        color_progress = ColorProgress(coloring=False)

        self.assertTrue(color_progress._do_update())

    @patch(" enaptorch.framework.color_formatting.jupyter", True)
    def test_do_update_with_jupyter_coloring_and_no_previous_timer_return_true(self):
        color_progress = ColorProgress(coloring=True)

        self.assertTrue(color_progress._do_update())

    @patch(" enaptorch.framework.color_formatting.jupyter", True)
    @patch(" enaptorch.framework.color_formatting.time")
    def test_do_update_with_jupyter_coloring_and_no_update_needed_return_false(self, time_mock):
        color_progress = ColorProgress(coloring=True)
        color_progress.prev_print_time = 1.0
        time_mock.time.return_value = 1.01

        self.assertFalse(color_progress._do_update())

    @patch(" enaptorch.framework.color_formatting.jupyter", True)
    @patch(" enaptorch.framework.color_formatting.time")
    def test_do_update_with_jupyter_coloring_and_an_update_needed_return_true(self, time_mock):
        color_progress = ColorProgress(coloring=True)
        color_progress.prev_print_time = 1.0
        time_mock.time.return_value = 1.2

        self.assertTrue(color_progress._do_update())



 
import os
from tempfile import TemporaryDirectory
from unittest import TestCase

from torch import nn
import torch.nn.functional as F

from  enaptorch import Experiment, acc, FBeta


class ExperimentTasksTest(TestCase):
    def setUp(self):
        self.temp_dir_obj = TemporaryDirectory()
        self.test_checkpoints_path = os.path.join(self.temp_dir_obj.name, 'expt')

    def tearDown(self):
        self.temp_dir_obj.cleanup()

    def test_task_classif(self):
        expt = Experiment(
            self.test_checkpoints_path,
            nn.Linear(1, 10),
            optimizer='sgd',
            task='classif',
        )
        self._test_task_classif(expt)

        expt = Experiment(
            self.test_checkpoints_path,
            nn.Linear(1, 10),
            optimizer='sgd',
            task='classification',
        )
        self._test_task_classif(expt)

    def _test_task_classif(self, expt):
        self.assertEqual(expt.monitor_metric, 'val_acc')
        self.assertEqual(expt.monitor_mode, 'max')
        self.assertEqual(expt.model.loss_function.func, F.cross_entropy)
        self.assertEqual(len(expt.model.batch_metrics), 1)
        self.assertEqual(expt.model.batch_metrics[0].func, acc)
        self.assertEqual(len(expt.model.epoch_metrics), 1)
        self.assertIsInstance(expt.model.epoch_metrics[0], FBeta)

    def test_task_regr(self):
        expt = Experiment(
            self.test_checkpoints_path,
            nn.Linear(1, 10),
            optimizer='sgd',
            task='regr',
        )
        self._test_task_regr(expt)

        expt = Experiment(
            self.test_checkpoints_path,
            nn.Linear(1, 10),
            optimizer='sgd',
            task='regression',
        )
        self._test_task_regr(expt)

    def _test_task_regr(self, expt):
        self.assertEqual(expt.monitor_metric, 'val_loss')
        self.assertEqual(expt.monitor_mode, 'min')
        self.assertEqual(expt.model.loss_function.func, F.mse_loss)
        self.assertEqual(len(expt.model.batch_metrics), 0)
        self.assertEqual(len(expt.model.epoch_metrics), 0)



 
import os
from typing import List, Dict
from typing import Mapping, Sequence
from unittest import TestCase
from unittest.mock import patch, MagicMock, call

import git
import torch
import torch.nn as nn
from mlflow.exceptions import MlflowException
from omegaconf import DictConfig

from  enaptorch import Model
from  enaptorch.framework.callbacks.mlflow_logger import _get_git_commit, MLFlowLogger
from tests.framework.tools import SomeDataGeneratorWithLen

a_git_commit = "9bff900c30e80c3a35388d3e617db5b7a64c9afd"
mlflow_default_git_commit_tag = "mlflow.source.git.commit"


class MLFlowLoggerTest(TestCase):
    def setUp(self) -> None:
        self.a_experiment_name = "a_name"
        self.a_run_id = "101010"
        self.a_experiment_id = "112321"
        self.none_tracking_uri = None

        self.a_exception_message = "a error message"

        self.the_working_directory = os.getcwd()

        self.experiment_mock = MagicMock()
        self.experiment_mock.return_value = self.a_experiment_id

        self.run_mock = MagicMock()
        self.run_mock.return_value.info.run_id = self.a_run_id

        self.num_epochs = 2
        self.steps_per_epoch = 5
        self.batch_size = 32
        self.a_log = {"metric_1": 1, "metric_2": 2}

        self.settings_in_dict = {"param_1": 1, "param_2": 2, "param_3": "value"}
        self.settings_in_dict_config_no_sequence = DictConfig(
            {'param_dict': {'param_1': 1}, 'param_dict_2': {'param_2"': 2, 'param_3"': 3}, 'param': 'value'}
        )
        self.settings_in_dict_config_with_sequence = DictConfig(
            {
                'param_dict': {'param_1': 1},
                'param_dict_2': {'param_2"': 2, 'param_3"': 3},
                'param': 'value',
                'a_list_param': [0, 1],
            }
        )

    @patch(" enaptorch.framework.mlflow_logger.mlflow", None)
    def test_whenMLFowNotInstalled_thenRaiseImportError(self):
        with self.assertRaises(ImportError):
            MLFlowLogger(self.a_experiment_name)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_whenNewExperiment_givenAMLFlowInstantiation_thenCreateNewExperiment(self):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = self.experiment_mock

            MLFlowLogger(self.a_experiment_name)

            create_experiment_call = [call().create_experiment(self.a_experiment_name, self.none_tracking_uri)]

            ml_flow_client_patch.assert_has_calls(create_experiment_call)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_whenExperimentAlreadyCreated_givenAMLFlowInstantiation_thenGetExperiment(self):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = MagicMock(
                side_effect=MlflowException(self.a_exception_message)
            )
            ml_flow_client_patch.return_value.get_experiment_by_name = MagicMock(return_value=self.experiment_mock)

            MLFlowLogger(self.a_experiment_name)

            create_experiment_calls = [
                call().create_experiment(self.a_experiment_name, self.none_tracking_uri),
                call().get_experiment_by_name(self.a_experiment_name),
            ]

            ml_flow_client_patch.assert_has_calls(create_experiment_calls)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_whenCorrectSettings_givenAMLFlowInstantiation_thenMLFlowClientIsProperlySet(self):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = self.experiment_mock
            ml_flow_client_patch.return_value.create_run = self.run_mock

            mlflow_logger = MLFlowLogger(self.a_experiment_name)

            settings_calls = [
                call().create_experiment(self.a_experiment_name, self.none_tracking_uri),
                call().create_run(experiment_id=self.a_experiment_id),
            ]
            ml_flow_client_patch.assert_has_calls(settings_calls)

            actual_experiment_id = mlflow_logger.experiment_id
            expected_experiment_id = self.a_experiment_id
            self.assertEqual(expected_experiment_id, actual_experiment_id)

            actual_run_id = mlflow_logger.run_id
            expected_run_id = self.a_run_id
            self.assertEqual(expected_run_id, actual_run_id)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", side_effect=[a_git_commit])
    def test_whenGitRepo_givenAMLFlowInstantiation_thenLogGitCommit(self, get_git_commit_patch):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_run = self.run_mock
            MLFlowLogger(self.a_experiment_name)

            git_logging_call = [call(self.the_working_directory)]
            get_git_commit_patch.assert_has_calls(git_logging_call)

            mlflow_client_call = [call().set_tag(self.a_run_id, mlflow_default_git_commit_tag, a_git_commit)]
            ml_flow_client_patch.assert_has_calls(mlflow_client_call)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_whenLogMetric_givenAMLFlowCallback_thenLogMetric(self):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = self.experiment_mock
            ml_flow_client_patch.return_value.create_run = self.run_mock

            mlflow_logger = MLFlowLogger(self.a_experiment_name)

            ml_flow_client_calls = []
            for key, value in self.a_log.items():
                mlflow_logger.log_metric(key, value)
                ml_flow_client_calls.append(call().log_metric(run_id=self.a_run_id, key=key, value=value, step=None))
            ml_flow_client_patch.assert_has_calls(ml_flow_client_calls)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_whenLogParam_givenAMLFlowCallback_thenLogParam(self):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = self.experiment_mock
            ml_flow_client_patch.return_value.create_run = self.run_mock

            mlflow_logger = MLFlowLogger(self.a_experiment_name)

            ml_flow_client_calls = []
            for key, value in self.settings_in_dict.items():
                mlflow_logger.log_param(key, value)
                ml_flow_client_calls.append(call().log_param(run_id=self.a_run_id, key=key, value=value))
            ml_flow_client_patch.assert_has_calls(ml_flow_client_calls)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_whenLogConfigParamsASimpleDict_givenAMLFlowCallback_thenLogParams(self):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = self.experiment_mock
            ml_flow_client_patch.return_value.create_run = self.run_mock

            mlflow_logger = MLFlowLogger(self.a_experiment_name)
            mlflow_logger.log_config_params(self.settings_in_dict)

            ml_flow_client_calls = []
            for key, value in self.settings_in_dict.items():
                ml_flow_client_calls.append(call().log_param(run_id=self.a_run_id, key=key, value=value))
            ml_flow_client_patch.assert_has_calls(ml_flow_client_calls)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_whenLogConfigParamsAConfigDict_givenAMLFlowCallback_thenLogParams(self):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = self.experiment_mock
            ml_flow_client_patch.return_value.create_run = self.run_mock

            mlflow_logger = MLFlowLogger(self.a_experiment_name)
            mlflow_logger.log_config_params(self.settings_in_dict_config_no_sequence)

            ml_flow_client_calls = self._populate_calls_from_dict(self.settings_in_dict_config_no_sequence)
            ml_flow_client_patch.assert_has_calls(ml_flow_client_calls)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_whenLogConfigParamsAConfigDictWithSequence_givenAMLFlowCallback_thenLogParams(self):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = self.experiment_mock
            ml_flow_client_patch.return_value.create_run = self.run_mock

            mlflow_logger = MLFlowLogger(self.a_experiment_name)
            mlflow_logger.log_config_params(self.settings_in_dict_config_with_sequence)

            ml_flow_client_calls = self._populate_calls_from_dict(self.settings_in_dict_config_with_sequence)
            ml_flow_client_patch.assert_has_calls(ml_flow_client_calls)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_whenOnTrainEndSuccess_givenAMLFlowCallback_thenLogLastEpochNumber(self):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = self.experiment_mock
            ml_flow_client_patch.return_value.create_run = self.run_mock

            mlflow_logger = MLFlowLogger(self.a_experiment_name)
            mlflow_logger.set_params({"epochs": self.num_epochs})
            mlflow_logger.on_train_end(self.a_log)

            ml_flow_client_calls = [
                call().log_metric(run_id=self.a_run_id, key='last-epoch', value=self.num_epochs, step=None)
            ]
            ml_flow_client_patch.assert_has_calls(ml_flow_client_calls)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_whenOnTrainEndSuccess_givenAMLFlowCallback_thenHasSuccessTerminatedStatus(self):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = self.experiment_mock
            ml_flow_client_patch.return_value.create_run = self.run_mock

            mlflow_logger = MLFlowLogger(self.a_experiment_name)
            mlflow_logger.set_params({"epochs": self.num_epochs})
            mlflow_logger.on_train_end(self.a_log)

            ml_flow_client_calls = [call().set_terminated(self.a_run_id, status="FINISHED")]
            ml_flow_client_patch.assert_has_calls(ml_flow_client_calls)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_whenOnTestFailure_givenAMLFlowCallback_thenHasFailureTerminatedStatus(self):
        # pylint: disable=protected-access
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = self.experiment_mock
            ml_flow_client_patch.return_value.create_run = self.run_mock

            mlflow_logger = MLFlowLogger(self.a_experiment_name)
            mlflow_logger._status_handling()

            ml_flow_client_calls = [call().set_terminated(self.a_run_id, status="FAILED")]
            ml_flow_client_patch.assert_has_calls(ml_flow_client_calls)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_whenOnTestSuccess_givenAMLFlowCallback_thenHasSuccessTerminatedStatus(self):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = self.experiment_mock
            ml_flow_client_patch.return_value.create_run = self.run_mock

            mlflow_logger = MLFlowLogger(self.a_experiment_name)
            mlflow_logger.set_params({"epochs": self.num_epochs})
            mlflow_logger.on_train_end(self.a_log)
            mlflow_logger.on_test_begin({})  # since we change status at the start of testing
            mlflow_logger.on_test_end(self.a_log)

            ml_flow_client_calls = [call().set_terminated(self.a_run_id, status="FINISHED")]
            ml_flow_client_patch.assert_has_calls(ml_flow_client_calls)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_integration_train(self):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = self.experiment_mock
            ml_flow_client_patch.return_value.create_run = self.run_mock

            mlflow_logger = MLFlowLogger(self.a_experiment_name)

            torch.manual_seed(42)
            a_pytorch_network = nn.Linear(1, 1)
            a_loss_function = nn.MSELoss()
            an_optimizer = torch.optim.Adam(a_pytorch_network.parameters(), lr=1e-3)
            model = Model(a_pytorch_network, an_optimizer, a_loss_function)

            train_generator = SomeDataGeneratorWithLen(self.batch_size, 10, 0)
            valid_generator = SomeDataGeneratorWithLen(self.batch_size, 10, 0)
            logs = model.fit_generator(
                train_generator,
                valid_generator,
                epochs=self.num_epochs,
                steps_per_epoch=self.steps_per_epoch,
                callbacks=[mlflow_logger],
            )

            ml_flow_client_calls = self._populate_calls_from_logs(logs)
            ml_flow_client_patch.assert_has_calls(ml_flow_client_calls)

    @patch(" enaptorch.framework.mlflow_logger._get_git_commit", MagicMock())
    def test_integration_train_batch_granularity(self):
        with patch(" enaptorch.framework.mlflow_logger.MlflowClient") as ml_flow_client_patch:
            ml_flow_client_patch.return_value.create_experiment = self.experiment_mock
            ml_flow_client_patch.return_value.create_run = self.run_mock

            mlflow_logger = MLFlowLogger(self.a_experiment_name, batch_granularity=True)

            torch.manual_seed(42)
            a_pytorch_network = nn.Linear(1, 1)
            a_loss_function = nn.MSELoss()
            an_optimizer = torch.optim.Adam(a_pytorch_network.parameters(), lr=1e-3)
            model = Model(a_pytorch_network, an_optimizer, a_loss_function)

            train_generator = SomeDataGeneratorWithLen(self.batch_size, 10, 0)
            valid_generator = SomeDataGeneratorWithLen(self.batch_size, 10, 0)
            a_num_epoch = 2
            model.fit_generator(
                train_generator,
                valid_generator,
                epochs=a_num_epoch,
                steps_per_epoch=self.steps_per_epoch,
                callbacks=[mlflow_logger],
            )

            self._assert_has_granularity_calls(ml_flow_client_patch)

    def _populate_calls_from_dict(self, config_dict: Dict) -> List:
        ml_flow_client_calls = []
        for key, value in config_dict.items():
            if isinstance(value, Mapping):
                for key_lower, value_lower in value.items():
                    good_key = f"{key}.{key_lower}"
                    ml_flow_client_calls.append(call().log_param(run_id=self.a_run_id, key=good_key, value=value_lower))
            elif isinstance(value, Sequence) and not isinstance(value, str):
                for idx, value_lower in enumerate(value):
                    good_key = f"{key}.{idx}"
                    ml_flow_client_calls.append(
                        (call().log_param(run_id=self.a_run_id, key=good_key, value=value_lower))
                    )
            else:
                ml_flow_client_calls.append(call().log_param(run_id=self.a_run_id, key=key, value=value))
        return ml_flow_client_calls

    def _populate_calls_from_logs(self, logs: Dict) -> List:
        ml_flow_client_calls = []
        for epoch_num, epoch_log in enumerate(logs):
            epoch_log.pop("epoch")
            for key, value in epoch_log.items():
                ml_flow_client_calls.append(
                    call().log_metric(run_id=self.a_run_id, key=key, value=value, step=epoch_num + 1)
                )  # +1 for enumerate
        ml_flow_client_calls.append(
            call().log_metric(run_id=self.a_run_id, key="last-epoch", value=self.num_epochs, step=None)
        )
        ml_flow_client_calls.append(call().set_terminated(self.a_run_id, status='FINISHED'))
        return ml_flow_client_calls

    def _assert_has_granularity_calls(self, ml_flow_client_patch):
        for _ in range(1, self.num_epochs):
            for step_number in range(1, self.steps_per_epoch):
                ml_flow_client_step_calls = []
                ml_flow_client_step_calls.append(
                    call().log_metric(run_id=self.a_run_id, key='batch', value=step_number, step=step_number)
                )
                ml_flow_client_step_calls.append(
                    call().log_metric(run_id=self.a_run_id, key='size', value=self.batch_size, step=step_number)
                )
                ml_flow_client_patch.assert_has_calls(ml_flow_client_step_calls)


class GetGitCommitTest(TestCase):
    def setUp(self) -> None:
        self.a_fake_path = "a_fake_path"
        self.a_wrong_path = "/a_wrong_path"
        self.a_git_sha = a_git_commit

    @patch(" enaptorch.framework.mlflow_logger.git", None)
    def test_whenGitNotInstall_givenARepositoryPathToGetGitCommit_thenRaiseWarning(self):
        with self.assertWarns(UserWarning):
            _get_git_commit(self.a_fake_path)

    @patch(" enaptorch.framework.mlflow_logger.git", None)
    def test_whenGitNotInstall_givenARepositoryPathToGetGitCommit_thenGitCommitIsNone(self):
        with self.assertWarns(UserWarning):
            commit = _get_git_commit(self.a_fake_path)
        self.assertIsNone(commit)

    @patch(" enaptorch.framework.mlflow_logger.git.Repo")
    def test_whenGitInstalled_givenARepositoryPathToGetGitCommitButNotAGitRepo_thenRaiseWarning(self, git_repo_patch):
        git_repo_patch.side_effect = git.NoSuchPathError()
        with self.assertWarns(UserWarning):
            _get_git_commit(self.a_fake_path)

    def test_whenGitInstalled_givenAWrongRepositoryPathToGetGitCommit_thenRaiseWarning(self):
        with self.assertWarns(UserWarning):
            _get_git_commit(self.a_wrong_path)

    @patch(" enaptorch.framework.mlflow_logger.git.Repo")
    def test_whenGitInstalled_givenARepositoryPathToGetGitCommitAndAGitRepo_thenReturnCommit(self, git_repo_patch):
        git_repo_patch.return_value.head.commit.hexsha = self.a_git_sha

        actual = _get_git_commit(self.a_fake_path)
        expected = self.a_git_sha

        self.assertEqual(expected, actual)



from abc import ABC, abstractmethod

import torch.nn as nn


class Metric(ABC, nn.Module):
    """
    The abstract class representing a metric which can be accumulated at each batch and calculated at the end
    of the epoch.
    """

    def forward(self, y_pred, y_true):
        """
        Update the current state of the metric and return the metric for the current batch. This method has to
        be implemented if the metric is used as a **batch metric**. If used as an epoch metric, it does not need to be
        implemented.

        Args:
            y_pred: The prediction of the model.
            y_true: Target to evaluate the model.

        Returns:
            The value of the metric for the current batch.
        """
        raise NotImplementedError

    def update(self, y_pred, y_true) -> None:
        """
        Update the current state of the metric. This method has to be implemented if the metric is used as an **epoch
        metric**. If used as a batch metric, it does not need to be implemented.

        Args:
            y_pred: The prediction of the model.
            y_true: Target to evaluate the model.
        """
        raise NotImplementedError

    @abstractmethod
    def compute(self):
        """
        Compute and return the metric. Should not modify the state of metric.
        """
        pass

    @abstractmethod
    def reset(self) -> None:
        """
        The information kept for the computation of the metric is cleaned so that a new epoch can be done.
        """
        pass



 
from unittest import TestCase
from unittest.mock import MagicMock

from  enaptorch import Callback, CallbackList


class CallbackListTest(TestCase):
    def setUp(self) -> None:
        self.initial_callback = MagicMock(spec=Callback)
        self.callback_list = CallbackList([self.initial_callback])

    def test_append_callback(self):
        self.assertEqual(len(self.callback_list.callbacks), 1)
        a_callback = MagicMock(spec=Callback)
        self.callback_list.append(a_callback)

        self.assertEqual(len(self.callback_list.callbacks), 2)

    def test_iterator(self):
        a_callback = MagicMock(spec=Callback)
        self.callback_list.append(a_callback)
        self.assertEqual(len(list(self.callback_list)), 2)



from typing import Dict

import numpy as np

from .callbacks import Callback


class EarlyStopping(Callback):
    """
    The source code of this class is under the MIT License and was copied from the Keras project,
    and has been modified.

    Stop training when a monitored quantity has stopped improving.

    Args:
        monitor (str): Quantity to be monitored.
        min_delta (float): Minimum change in the monitored quantity to qualify as an improvement,
            i.e. an absolute change of less than min_delta, will count as no improvement.
            (Default value = 0)
        patience (int): Number of epochs with no improvement after which training will be stopped.
            (Default value = 0)
        verbose (bool): Whether to print when early stopping is done.
            (Default value = False)
        mode (str): One of {'min', 'max'}. In `min` mode, training will stop when the quantity
            monitored has stopped decreasing; in `max` mode it will stop when the quantity monitored has
            stopped increasing.
            (Default value = 'min')
    """

    def __init__(
        self,
        *,
        monitor: str = 'val_loss',
        min_delta: float = 0.0,
        patience: int = 0,
        verbose: bool = False,
        mode: str = 'min',
    ):
        super().__init__()

        self.monitor = monitor
        self.patience = patience
        self.verbose = verbose
        self.min_delta = min_delta
        self.wait = 0
        self.stopped_epoch = 0

        if mode not in ['min', 'max']:
            raise ValueError(f"Invalid mode '{mode}'")
        self.mode = mode

        if mode == 'min':
            self.min_delta *= -1
            self.monitor_op = np.less
        elif mode == 'max':
            self.min_delta *= 1
            self.monitor_op = np.greater

    def on_train_begin(self, logs: Dict):
        # Allow instances to be re-used
        self.wait = 0
        self.stopped_epoch = 0
        self.best = np.Inf if self.mode == 'min' else -np.Inf

    def on_epoch_end(self, epoch_number: int, logs: Dict):
        current = logs[self.monitor]
        if self.monitor_op(current - self.min_delta, self.best):
            self.best = current
            self.wait = 0
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_epoch = epoch_number
                self.model.stop_training = True
                if self.verbose:
                    print(f'Epoch {self.stopped_epoch}: early stopping')



from typing import Dict, List


class Callback:
    # pylint: disable=too-many-public-methods
    """
    Attributes:
        params (dict): Contains ``'epoch'`` and ``'steps_per_epoch'`` keys which are passed to the
            when training. Contains ``'steps'`` when evaluating. May contain other keys.
        model (Model): A reference to the :class:`~ enaptorch.Model` object which is using the callback.
    """

    def __init__(self):
        self.model = None
        self.params = None

    def set_params(self, params: Dict):
        self.params = params

    def set_model(self, model):
        self.model = model

    def on_epoch_begin(self, epoch_number: int, logs: Dict):
        """
        Is called before the beginning of each epoch.

        Args:
            epoch_number (int): The epoch number.
            logs (dict): Usually an empty dict.
        """
        pass

    def on_epoch_end(self, epoch_number: int, logs: Dict):
        """
        Is called before the end of each epoch.

        Args:
            epoch_number (int): The epoch number.
            logs (dict): Contains the following keys:

                 * ``'epoch'``: The epoch number.
                 * ``'time'``: The computation time of the epoch.
                 * ``'loss'``: The average loss of the batches.
                 * Values of training metrics: One key for each type of metrics. The metrics are also averaged.
                 * ``'val_loss'``: The average loss of the batches on the validation set.
                 * Values of validation metrics: One key for each type of metrics on the validation set. Each key is
                   prefixed by ``'val_'``. The metrics are also averaged.

        Example::

            logs = {'epoch': 2, 'time': 6.08248, 'loss': 0.40161, 'acc': 89.052, 'fscore_micro': 0.89051,
                    'val_loss': 0.36814, 'val_acc': 89.52, 'val_fscore_micro': 0.89520}
        """
        pass

    def on_train_batch_begin(self, batch_number: int, logs: Dict):
        """
        Is called before the beginning of the training batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Usually an empty dict.
        """
        pass

    def on_train_batch_end(self, batch_number: int, logs: Dict):
        """
        Is called before the end of the training batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Contains the following keys:

                 * ``'batch'``: The batch number.
                 * ``'size'``: The size of the batch as inferred by :func:`~Model.get_batch_size()`.
                 * ``'time'``: The computation time of the batch.
                 * ``'loss'``: The loss of the batch.
                 * Values of the batch metrics for the specific batch: One key for each type of metrics.

        Example::

            logs = {'batch': 171, 'size': 32, 'time': 0.00310, 'loss': 1.95204, 'acc': 43.75}
        """
        pass

    def on_valid_batch_begin(self, batch_number: int, logs: Dict):
        """
        Is called before the beginning of the validation batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Usually an empty dict.
        """
        pass

    def on_valid_batch_end(self, batch_number: int, logs: Dict):
        """
        Is called before the end of the validation batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Contains the following keys:

                 * ``'batch'``: The batch number.
                 * ``'size'``: The size of the batch as inferred by :func:`~Model.get_batch_size()`.
                 * ``'time'``: The computation time of the batch.
                 * ``val_loss'``: The loss of the batch.
                 * Values of the batch metrics for the specific batch: One key for each type of metrics. Each key is
                   prefixed by ``'val_'``. The metrics are also averaged.

        Example::

            logs = {'batch': 171, 'size': 32, 'time': 0.00310, 'val_loss': 1.95204, 'val_acc': 43.75}
        """
        pass

    def on_test_batch_begin(self, batch_number: int, logs: Dict):
        """
        Is called before the beginning of the testing batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Usually an empty dict.
        """
        pass

    def on_test_batch_end(self, batch_number: int, logs: Dict):
        """
        Is called before the end of the testing batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Contains the following keys:

                 * ``'batch'``: The batch number.
                 * ``'size'``: The size of the batch as inferred by :func:`~Model.get_batch_size()`.
                 * ``'time'``: The computation time of the batch.
                 * ``'loss'``: The loss of the batch.
                 * Values of the batch metrics for the specific batch: One key for each type of metrics. Each key is
                   prefixed by ``'test_'``. The metrics are also averaged.

        Example::

            logs = {'batch': 171, 'size': 32, 'time': 0.00310, 'test_loss': 1.95204, 'test_acc': 43.75}
        """
        pass

    def on_predict_batch_begin(self, batch_number: int, logs: Dict):
        """
        Is called before the beginning of the predict batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Usually an empty dict.
        """
        pass

    def on_predict_batch_end(self, batch_number: int, logs: Dict):
        """
        Is called before the end of the predict batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Contains the following keys:

                 * ``'batch'``: The batch number.
                 * ``'time'``: The computation time of the batch.

        Example::

            logs = {'batch': 171, 'time': 0.00310}
        """
        pass

    def on_train_begin(self, logs: Dict):
        """
        Is called before the beginning of the training.

        Args:
            logs (dict): Usually an empty dict.
        """
        pass

    def on_train_end(self, logs: Dict):
        """
        Is called before the end of the training.

        Args:
            logs (dict): Usually an empty dict.
        """
        pass

    def on_valid_begin(self, logs: Dict):
        """
        Is called before the beginning of the validation.

        Args:
            logs (dict): Usually an empty dict.
        """
        pass

    def on_valid_end(self, logs: Dict):
        """
        Is called before the end of the validation.

        Args:
            logs (dict): Contains the following keys:

                 * ``'time'``: The total computation time of the test.
                 * ``'val_loss'``: The average loss of the batches on the test set.
                 * Values of testing metrics: One key for each type of metrics. Each key is
                   prefixed by ``'val_'``. The metrics are also averaged.

        Example::

            logs = {'time': 6.08248, 'val_loss': 0.40161, 'val_acc': 89.052, 'val_fscore_micro': 0.89051}
        """
        pass

    def on_test_begin(self, logs: Dict):
        """
        Is called before the beginning of the testing.

        Args:
            logs (dict): Usually an empty dict.
        """
        pass

    def on_test_end(self, logs: Dict):
        """
        Is called before the end of the testing.

        Args:
            logs (dict): Contains the following keys:

                 * ``'time'``: The total computation time of the test.
                 * ``'test_loss'``: The average loss of the batches on the test set.
                 * Values of testing metrics: One key for each type of metrics. Each key is
                   prefixed by ``'test_'``. The metrics are also averaged.

        Example::

            logs = {'time': 6.08248, 'test_loss': 0.40161, 'test_acc': 89.052, 'test_fscore_micro': 0.89051}
        """
        pass

    def on_predict_begin(self, logs: Dict):
        """
        Is called before the beginning of the predict.

        Args:
            logs (dict): Usually an empty dict.
        """
        pass

    def on_predict_end(self, logs: Dict):
        """
        Is called before the end of the predict.

        Args:
            logs (dict): Contains the following keys:

                 * ``'time'``: The total computation time of the predict.

        Example::

            logs = {'time': 6.08248}
        """
        pass

    def on_backward_end(self, batch_number: int):
        """
        Is called after the backpropagation but before the optimization step.

        Args:
            batch_number (int): The batch number.
        """
        pass


class CallbackList:
    # pylint: disable=too-many-public-methods
    def __init__(self, callbacks: List[Callback]):
        callbacks = callbacks or []
        self.callbacks = list(callbacks)

    def append(self, callback: Callback):
        self.callbacks.append(callback)

    def set_params(self, params: Dict):
        for callback in self.callbacks:
            callback.set_params(params)

    def set_model(self, model):
        for callback in self.callbacks:
            callback.set_model(model)

    def on_epoch_begin(self, epoch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_epoch_begin(epoch_number, logs)

    def on_epoch_end(self, epoch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_epoch_end(epoch_number, logs)

    def on_train_batch_begin(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_train_batch_begin(batch_number, logs)

    def on_train_batch_end(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_train_batch_end(batch_number, logs)

    def on_valid_batch_begin(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_valid_batch_begin(batch_number, logs)

    def on_valid_batch_end(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_valid_batch_end(batch_number, logs)

    def on_test_batch_begin(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_test_batch_begin(batch_number, logs)

    def on_test_batch_end(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_test_batch_end(batch_number, logs)

    def on_predict_batch_begin(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_predict_batch_begin(batch_number, logs)

    def on_predict_batch_end(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_predict_batch_end(batch_number, logs)

    def on_train_begin(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_train_begin(logs)

    def on_train_end(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_train_end(logs)

    def on_valid_begin(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_valid_begin(logs)

    def on_valid_end(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_valid_end(logs)

    def on_test_begin(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_test_begin(logs)

    def on_test_end(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_test_end(logs)

    def on_predict_begin(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_predict_begin(logs)

    def on_predict_end(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_predict_end(logs)

    def on_backward_end(self, batch_number: int):
        for callback in self.callbacks:
            callback.on_backward_end(batch_number)

    def __iter__(self):
        return iter(self.callbacks)



from setuptools import setup, find_packages
import os

current_file_path = os.path.abspath(os.path.dirname(__file__))

def get_readme():
    readme_file_path = os.path.join(current_file_path, 'README.md')
    with open(readme_file_path, 'r', encoding='utf-8') as f:
        return f.read()

setup(
    name='enaptorch',
    version="v1.0",
    author='Swanand Katdare',
    author_email='swanand@edgeneural.ai',
    url='https://enaptorch.org',
    download_url='https://github.com/SwanandkEN/enaptorch/archive/refs/tags/v1.0.zip',
    packages=find_packages(),
    classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Intended Audience :: Education',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
        ],
        install_requires=['numpy', 'torch', 'torchmetrics'],
        python_requires='>=3.6.1',
        description='A simplified framework and utilities for PyTorch.',
        long_description=get_readme(),
        long_description_content_type='text/plain',
        extras_require={
            "colorama": "colorama>=0.4.3",
            "scikit-learn": "scikit-learn>=0.23.2",
            "tensorboard": "tensorboard>=2.4.0",
            "tensorboardX": "tensorboardX>=2.1",
            "torchvision": "torchvision>=0.8.1",
            "pandas": "pandas>=2.0.0.0",
            "mlflow": "mlflow>=1.12.1",
        },

)


from .utils import get_names_of_metric
from .metrics_registering import get_metric
from .base import Metric
from .decomposable import convert_decomposable_metric_to_object


class IndexedArgumentMetric(Metric):
    def __init__(self, metric, *, index=None, pred_index=None, true_index=None):
        super().__init__()

        if isinstance(metric, str):
            metric = get_metric(metric)
        names, metric = get_names_of_metric(metric)
        metric = convert_decomposable_metric_to_object(metric, names)

        self.__name__ = names
        self.metric = metric

        if index is not None and (pred_index is not None or true_index is not None):
            raise ValueError(
                "'pred_index' and 'true_index' arguments should should not be used " "with the 'index argument."
            )
        if index is not None:
            self.pred_index = index
            self.true_index = index
        else:
            self.pred_index = pred_index
            self.true_index = true_index

    def _select_indices(self, y_pred, y_true):
        if self.pred_index is not None:
            y_pred = y_pred[self.pred_index]
        if self.true_index is not None:
            y_true = y_true[self.true_index]
        return y_pred, y_true

    def forward(self, y_pred, y_true):
        y_pred, y_true = self._select_indices(y_pred, y_true)
        return self.metric(y_pred, y_true)

    def update(self, y_pred, y_true):
        y_pred, y_true = self._select_indices(y_pred, y_true)
        return self.metric.update(y_pred, y_true)

    def compute(self):
        return self.metric.compute()

    def reset(self) -> None:
        return self.metric.reset()



"""
The source code of this file was copied from the AllenNLP project, and has been modified. All modifications
made from the original source code are under the LGPLv3 license.


Copyright (c) 2022  enaptorch and all respective contributors.

Each contributor holds copyright over their respective contributions. The project versioning (Git)
records all such contribution source information on the  enaptorch and AllenNLP repository.

This file is part of  enaptorch.

 enaptorch is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
version.

 enaptorch is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with  enaptorch. If not, see
<https://www.gnu.org/licenses/>.


Copyright 2019 AllenNLP

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=protected-access
from unittest import TestCase

import numpy
import torch
import torch.nn as nn

from  enaptorch import FBeta, F1, BinaryF1, Model


class FBetaTest(TestCase):
    def setUp(self):
        # [0, 1, 1, 1, 3, 1]
        self.predictions = torch.Tensor(
            [
                [0.35, 0.25, 0.1, 0.1, 0.2],
                [0.1, 0.6, 0.1, 0.2, 0.0],
                [0.1, 0.6, 0.1, 0.2, 0.0],
                [0.1, 0.5, 0.1, 0.2, 0.0],
                [0.1, 0.2, 0.1, 0.7, 0.0],
                [0.1, 0.6, 0.1, 0.2, 0.0],
            ]
        )
        self.targets = torch.Tensor([0, 4, 1, 0, 3, 0])

        # detailed target state
        self.pred_sum = [1, 4, 0, 1, 0]
        self.true_sum = [3, 1, 0, 1, 1]
        self.true_positive_sum = [1, 1, 0, 1, 0]
        self.true_negative_sum = [3, 2, 6, 5, 5]
        self.total_sum = [6, 6, 6, 6, 6]

        desired_precisions = [1.00, 0.25, 0.00, 1.00, 0.00]
        desired_recalls = [0.33, 1.00, 0.00, 1.00, 0.00]
        desired_fscores = [
            (2 * p * r) / (p + r) if p + r != 0.0 else 0.0 for p, r in zip(desired_precisions, desired_recalls)
        ]
        self.desired_precisions = desired_precisions
        self.desired_recalls = desired_recalls
        self.desired_fscores = desired_fscores

    def test_config_errors(self):
        # Bad beta
        self.assertRaises(ValueError, FBeta, beta=0.0)

        # Bad average option
        self.assertRaises(ValueError, FBeta, average='mega')

        # F1 classes with beta different than 1.
        self.assertRaises(ValueError, F1, beta=2.0)
        self.assertRaises(ValueError, BinaryF1, beta=2.0)

        # Precision and recall with beta different than 1.
        self.assertWarns(UserWarning, FBeta, metric='precision', beta=2.0)
        self.assertWarns(UserWarning, FBeta, metric='recall', beta=2.0)

    def test_runtime_errors(self):
        fbeta = FBeta()
        # Metric was never called.
        self.assertRaises(RuntimeError, fbeta.compute)

    def test_fbeta_multiclass_state(self):
        fbeta = FBeta()
        fbeta.update(self.predictions, self.targets)

        # check state
        numpy.testing.assert_almost_equal(fbeta._pred_sum.tolist(), self.pred_sum)
        numpy.testing.assert_almost_equal(fbeta._true_sum.tolist(), self.true_sum)
        numpy.testing.assert_almost_equal(fbeta._true_positive_sum.tolist(), self.true_positive_sum)
        numpy.testing.assert_almost_equal(fbeta._total_sum.tolist(), self.total_sum)

    def test_fbeta_multiclass_with_mask(self):
        mask = torch.Tensor([1, 1, 1, 1, 1, 0])

        fbeta = FBeta()
        fbeta.update(self.predictions, (self.targets, mask))

        numpy.testing.assert_almost_equal(fbeta._pred_sum.tolist(), [1, 3, 0, 1, 0])
        numpy.testing.assert_almost_equal(fbeta._true_sum.tolist(), [2, 1, 0, 1, 1])
        numpy.testing.assert_almost_equal(fbeta._true_positive_sum.tolist(), [1, 1, 0, 1, 0])

    def test_fbeta_multiclass_with_ignore_index(self):
        targets = self.targets.clone()
        targets[-1] = -100
        fbeta = FBeta()
        fbeta.update(self.predictions, targets)

        numpy.testing.assert_almost_equal(fbeta._pred_sum.tolist(), [1, 3, 0, 1, 0])
        numpy.testing.assert_almost_equal(fbeta._true_sum.tolist(), [2, 1, 0, 1, 1])
        numpy.testing.assert_almost_equal(fbeta._true_positive_sum.tolist(), [1, 1, 0, 1, 0])

    def test_fbeta_multiclass_with_mask_and_ignore_index(self):
        targets = self.targets.clone()
        targets[-1] = -100
        mask = torch.Tensor([1, 1, 1, 1, 0, 1])

        fbeta = FBeta()
        fbeta.update(self.predictions, (targets, mask))

        numpy.testing.assert_almost_equal(fbeta._pred_sum.tolist(), [1, 3, 0, 0, 0])
        numpy.testing.assert_almost_equal(fbeta._true_sum.tolist(), [2, 1, 0, 0, 1])
        numpy.testing.assert_almost_equal(fbeta._true_positive_sum.tolist(), [1, 1, 0, 0, 0])

    def test_fbeta_multiclass_macro_average_metric(self):
        precision = self._compute(metric='precision', average='macro')
        recall = self._compute(metric='recall', average='macro')
        fscore = self._compute(metric='fscore', average='macro')

        macro_precision = numpy.mean(self.desired_precisions)
        macro_recall = numpy.mean(self.desired_recalls)
        macro_fscore = numpy.mean(self.desired_fscores)

        # check type
        assert isinstance(precision, float)
        assert isinstance(recall, float)
        assert isinstance(fscore, float)

        # check value
        numpy.testing.assert_almost_equal(precision, macro_precision, decimal=2)
        numpy.testing.assert_almost_equal(recall, macro_recall, decimal=2)
        numpy.testing.assert_almost_equal(fscore, macro_fscore, decimal=2)

    def test_fbeta_multiclass_micro_average_metric(self):
        precision = self._compute(metric='precision', average='micro')
        recall = self._compute(metric='recall', average='micro')
        fscore = self._compute(metric='fscore', average='micro')

        true_positives = [1, 1, 0, 1, 0]
        false_positives = [0, 3, 0, 0, 0]
        false_negatives = [2, 0, 0, 0, 1]
        mean_true_positive = numpy.mean(true_positives)
        mean_false_positive = numpy.mean(false_positives)
        mean_false_negative = numpy.mean(false_negatives)

        micro_precision = mean_true_positive / (mean_true_positive + mean_false_positive)
        micro_recall = mean_true_positive / (mean_true_positive + mean_false_negative)
        micro_fscore = (2 * micro_precision * micro_recall) / (micro_precision + micro_recall)
        # check value
        numpy.testing.assert_almost_equal(precision, micro_precision, decimal=2)
        numpy.testing.assert_almost_equal(recall, micro_recall, decimal=2)
        numpy.testing.assert_almost_equal(fscore, micro_fscore, decimal=2)

    def test_fbeta_multiclass_with_explicit_label(self):
        # same prediction but with and explicit label
        label = 3
        precision = self._compute(metric='precision', average=label)
        recall = self._compute(metric='recall', average=label)
        fscore = self._compute(metric='fscore', average=label)

        desired_precision = self.desired_precisions[label]
        desired_recall = self.desired_recalls[label]
        desired_fscore = self.desired_fscores[label]
        # check value
        numpy.testing.assert_almost_equal(precision, desired_precision, decimal=2)
        numpy.testing.assert_almost_equal(recall, desired_recall, decimal=2)
        numpy.testing.assert_almost_equal(fscore, desired_fscore, decimal=2)

    def test_fbeta_multiclass_macro_average_metric_multireturn(self):
        fbeta = FBeta(average='macro')
        fbeta.update(self.predictions, self.targets)
        fscore, precision, recall = fbeta.compute()

        macro_precision = numpy.mean(self.desired_precisions)
        macro_recall = numpy.mean(self.desired_recalls)
        macro_fscore = numpy.mean(self.desired_fscores)

        # check type
        assert isinstance(precision, float)
        assert isinstance(recall, float)
        assert isinstance(fscore, float)

        # check value
        numpy.testing.assert_almost_equal(precision, macro_precision, decimal=2)
        numpy.testing.assert_almost_equal(recall, macro_recall, decimal=2)
        numpy.testing.assert_almost_equal(fscore, macro_fscore, decimal=2)

    def test_fbeta_handles_batch_size_of_one(self):
        predictions = torch.Tensor([[0.2862, 0.3479, 0.1627, 0.2033]])
        targets = torch.Tensor([1])
        mask = torch.Tensor([1])

        fbeta = FBeta()
        fbeta.update(predictions, (targets, mask))

        numpy.testing.assert_almost_equal(fbeta._pred_sum.tolist(), [0.0, 1.0, 0.0, 0.0])
        numpy.testing.assert_almost_equal(fbeta._true_sum.tolist(), [0.0, 1.0, 0.0, 0.0])
        numpy.testing.assert_almost_equal(fbeta._true_positive_sum.tolist(), [0.0, 1.0, 0.0, 0.0])
        numpy.testing.assert_almost_equal(fbeta._total_sum.tolist(), [1.0, 1.0, 1.0, 1.0])

    def test_fbeta_with_return_batch_value(self):
        targets = self.targets.clone()
        targets[-1] = -100
        mask = torch.Tensor([1, 1, 1, 1, 0, 1])

        fbeta = FBeta()
        batch_value = fbeta(self.predictions, (targets, mask))
        epoch_value = fbeta.compute()

        self.assertEqual(batch_value, epoch_value)

    def _compute(self, *args, **kwargs):
        fbeta = FBeta(*args, **kwargs)
        fbeta.update(self.predictions, self.targets)
        return fbeta.compute()

    def test_names(self):
        fbeta = FBeta(average='macro')
        self.assertEqual(['fscore_macro', 'precision_macro', 'recall_macro'], fbeta.__name__)
        fbeta = FBeta(average='micro')
        self.assertEqual(['fscore_micro', 'precision_micro', 'recall_micro'], fbeta.__name__)
        fbeta = FBeta(average='micro', names=['f', 'p', 'r'])
        self.assertEqual(['f', 'p', 'r'], fbeta.__name__)
        fbeta = FBeta(average=0)
        self.assertEqual(['fscore_0', 'precision_0', 'recall_0'], fbeta.__name__)
        fbeta = FBeta(metric='fscore', average='macro')
        self.assertEqual('fscore_macro', fbeta.__name__)
        fbeta = FBeta(metric='fscore', average='micro')
        self.assertEqual('fscore_micro', fbeta.__name__)
        fbeta = FBeta(metric='fscore', average=0)
        self.assertEqual('fscore_0', fbeta.__name__)
        fbeta = FBeta(metric='precision', average='macro')
        self.assertEqual('precision_macro', fbeta.__name__)
        fbeta = FBeta(metric='precision', average='micro')
        self.assertEqual('precision_micro', fbeta.__name__)
        fbeta = FBeta(metric='precision', average=0)
        self.assertEqual('precision_0', fbeta.__name__)
        fbeta = FBeta(metric='recall', average='macro')
        self.assertEqual('recall_macro', fbeta.__name__)
        fbeta = FBeta(metric='recall', average='micro')
        self.assertEqual('recall_micro', fbeta.__name__)
        fbeta = FBeta(metric='recall', average=0)
        self.assertEqual('recall_0', fbeta.__name__)
        fbeta = FBeta(metric='fscore', average='macro', names='f')
        self.assertEqual('f', fbeta.__name__)
        fbeta = FBeta(average='macro', names=['f', "p", "r"])
        self.assertEqual(["f", "p", "r"], fbeta.__name__)
        fbeta = FBeta(average='binary')
        self.assertEqual(['fscore_binary_1', 'precision_binary_1', 'recall_binary_1'], fbeta.__name__)
        fbeta = FBeta(average='binary', pos_label=0)
        self.assertEqual(['fscore_binary_0', 'precision_binary_0', 'recall_binary_0'], fbeta.__name__)

    def test_predefined_names(self):
        epoch_metrics = [
            'f1',
            'precision',
            'recall',
            'binaryf1',
            'binf1',
            'binaryprecision',
            'binprecision',
            'binaryrecall',
            'binrecall',
        ]
        fmetric = ['fscore', 'precision', 'recall', 'fscore', 'fscore', 'precision', 'precision', 'recall', 'recall']
        average = ['macro', 'macro', 'macro', 'binary', 'binary', 'binary', 'binary', 'binary', 'binary']
        names = [
            'fscore_macro',
            'precision_macro',
            'recall_macro',
            'bin_fscore1',
            'bin_fscore2',
            'bin_precision1',
            'bin_precision2',
            'bin_recall1',
            'bin_recall2',
        ]
        model = Model(nn.Linear(10, 2), 'sgd', 'cross_entropy', epoch_metrics=epoch_metrics)
        actual_fmetric = [epoch_metric._metric for epoch_metric in model.epoch_metrics]
        actual_average = [epoch_metric._average for epoch_metric in model.epoch_metrics]
        self.assertEqual(fmetric, actual_fmetric)
        self.assertEqual(average, actual_average)
        self.assertEqual(names, model.epoch_metrics_names)


class FBetaBinaryTest(TestCase):
    def setUp(self):
        # [0, 1, 1, 1, 0, 1]
        self.predictions = torch.Tensor([[0.35, 0.25], [0.1, 0.6], [0.1, 0.6], [0.1, 0.5], [0.2, 0.1], [0.1, 0.6]])
        self.targets = torch.Tensor([0, 0, 1, 0, 1, 0])

        # detailed target state
        self.pred_sum = [2, 4]
        self.true_sum = [4, 2]
        self.true_positive_sum = [1, 1]
        self.true_negative_sum = [1, 1]
        self.total_sum = [6, 6]

        desired_precision = 0.25
        desired_recall = 0.5
        desired_fscore = (
            (2 * desired_precision * desired_recall) / (desired_precision + desired_recall)
            if desired_precision + desired_recall != 0.0
            else 0.0
        )
        self.output = [desired_fscore, desired_precision, desired_recall]

    def test_fbeta_binary(self):
        fbeta = FBeta(average='binary')
        fbeta.update(self.predictions, self.targets)

        # check state
        numpy.testing.assert_almost_equal(fbeta._pred_sum.tolist(), self.pred_sum)
        numpy.testing.assert_almost_equal(fbeta._true_sum.tolist(), self.true_sum)
        numpy.testing.assert_almost_equal(fbeta._true_positive_sum.tolist(), self.true_positive_sum)
        numpy.testing.assert_almost_equal(fbeta._total_sum.tolist(), self.total_sum)
        numpy.testing.assert_almost_equal(fbeta.compute(), self.output)

    def test_fbeta_binary_one_dim_pred(self):
        fbeta = FBeta(average='binary')
        fbeta.update(self.predictions[:, 1:] - self.predictions[:, 0:1], self.targets)

        # check state
        numpy.testing.assert_almost_equal(fbeta._pred_sum.tolist(), self.pred_sum)
        numpy.testing.assert_almost_equal(fbeta._true_sum.tolist(), self.true_sum)
        numpy.testing.assert_almost_equal(fbeta._true_positive_sum.tolist(), self.true_positive_sum)
        numpy.testing.assert_almost_equal(fbeta._total_sum.tolist(), self.total_sum)
        numpy.testing.assert_almost_equal(fbeta.compute(), self.output)

    def test_fbeta_binary_zero_dim_pred(self):
        fbeta = FBeta(average='binary')
        fbeta.update(self.predictions[:, 1] - self.predictions[:, 0], self.targets)

        # check state
        numpy.testing.assert_almost_equal(fbeta._pred_sum.tolist(), self.pred_sum)
        numpy.testing.assert_almost_equal(fbeta._true_sum.tolist(), self.true_sum)
        numpy.testing.assert_almost_equal(fbeta._true_positive_sum.tolist(), self.true_positive_sum)
        numpy.testing.assert_almost_equal(fbeta._total_sum.tolist(), self.total_sum)
        numpy.testing.assert_almost_equal(fbeta.compute(), self.output)

    def test_fbeta_binary_with_return_batch_value(self):
        fbeta = FBeta(average='binary')
        batch_value = fbeta(self.predictions, self.targets)
        epoch_value = fbeta.compute()

        self.assertEqual(batch_value, epoch_value)



from .metrics_registering import *
from .pytorch_loss_functions_registering import *
from .epoch_metrics import *
from .decomposable import *
from .predefined import *
from .utils import *
from .metric_argument_indexing import *



 
# pylint: disable=unused-argument

# Bug with PyTorch source code makes torch.tensor as not callable for pylint.
# pylint: disable=not-callable
import os
import unittest
from unittest import skipIf
from unittest.mock import ANY
import numpy as np
import torch
import torch.nn as nn
import torchmetrics

from  enaptorch import Model, EpochMetric, rename_doubles, do_register_metric_class, unregister_metric_class


class MyConstTorchMetric(torchmetrics.Metric):
    def __init__(self, value=0):
        super().__init__()
        self.value = value

    def update(self, preds: torch.Tensor, target: torch.Tensor):
        pass

    def compute(self):
        return self.value


class ConstEpochMetric(EpochMetric):
    def __init__(self, value=0):
        super().__init__()
        self.value = value

    def forward(self, y_pred, y_true):
        pass

    def get_metric(self):
        return self.value

    def reset(self):
        pass


def get_batch_metric(value):
    def some_metric_name(y_pred, y_true):
        return torch.FloatTensor([value])

    return some_metric_name


class SomeMetricName(ConstEpochMetric):
    def get_metric(self):
        return torch.FloatTensor([self.value])

    def reset(self):
        pass


def get_const_batch_metric(value):
    def const_batch_metric(y_pred, y_true):
        return value

    return const_batch_metric


class MetricsModelIntegrationTest(unittest.TestCase):
    # pylint: disable=too-many-public-methods
    epochs = 2
    steps_per_epoch = 3
    batch_size = 10

    cuda_device = int(os.environ.get('CUDA_DEVICE', 0))

    def setUp(self):
        torch.manual_seed(42)
        self.pytorch_network = nn.Linear(1, 1)
        self.loss_function = nn.MSELoss()
        self.optimizer = torch.optim.SGD(self.pytorch_network.parameters(), lr=1e-3)
        dataset_size = MetricsModelIntegrationTest.batch_size * MetricsModelIntegrationTest.steps_per_epoch
        self.train_x = torch.rand(dataset_size, 1)
        self.train_y = torch.rand(dataset_size, 1)
        self.valid_x = torch.rand(dataset_size, 1)
        self.valid_y = torch.rand(dataset_size, 1)

        self.metric_names = ('a', 'b')
        self.metric_values = [1, 2]

    def test_repeated_batch_metrics_handling(self):
        expected_names = ['some_metric_name1', 'some_metric_name2']
        model = Model(
            self.pytorch_network,
            self.optimizer,
            self.loss_function,
            batch_metrics=[get_batch_metric(1), get_batch_metric(2)],
        )
        self._test_history(model, expected_names, [1, 2])

    def test_repeated_epoch_metrics_handling(self):
        expected_names = ['some_metric_name1', 'some_metric_name2']
        model = Model(
            self.pytorch_network,
            self.optimizer,
            self.loss_function,
            epoch_metrics=[SomeMetricName(1), SomeMetricName(2)],
        )
        self._test_history(model, expected_names, [1, 2])

    def test_repeated_batch_epoch_metrics_handling(self):
        expected_names = ['some_metric_name1', 'some_metric_name2']
        model = Model(
            self.pytorch_network,
            self.optimizer,
            self.loss_function,
            batch_metrics=[get_batch_metric(1)],
            epoch_metrics=[SomeMetricName(2)],
        )
        self._test_history(model, expected_names, [1, 2])

    def test_batch_metrics_with_multiple_names_returned_by_dict(self):
        d = dict(zip(self.metric_names, self.metric_values))
        batch_metric = get_const_batch_metric(d)
        model = Model(
            self.pytorch_network, self.optimizer, self.loss_function, batch_metrics=[(self.metric_names, batch_metric)]
        )
        self._test_history(model, d.keys(), d.values())

    def test_epoch_metrics_with_multiple_names_returned_by_dict(self):
        d = dict(zip(self.metric_names, self.metric_values))
        epoch_metric = ConstEpochMetric(d)
        model = Model(
            self.pytorch_network, self.optimizer, self.loss_function, epoch_metrics=[(self.metric_names, epoch_metric)]
        )
        self._test_history(model, d.keys(), d.values())

    def test_batch_metrics_with_multiple_names_returned_by_tensor(self):
        batch_metric = get_const_batch_metric(torch.tensor(self.metric_values))
        model = Model(
            self.pytorch_network, self.optimizer, self.loss_function, batch_metrics=[(self.metric_names, batch_metric)]
        )
        self._test_history(model, self.metric_names, self.metric_values)

    def test_epoch_metrics_with_multiple_names_returned_by_tensor(self):
        epoch_metric = ConstEpochMetric(torch.tensor(self.metric_values))
        model = Model(
            self.pytorch_network, self.optimizer, self.loss_function, epoch_metrics=[(self.metric_names, epoch_metric)]
        )
        self._test_history(model, self.metric_names, self.metric_values)

    def test_epoch_metrics_with_name_with_multiple_names_returned_by_tensor(self):
        class EpochMetricWithName(ConstEpochMetric):
            __name__ = self.metric_names

        epoch_metric = EpochMetricWithName(torch.tensor(self.metric_values))
        model = Model(self.pytorch_network, self.optimizer, self.loss_function, epoch_metrics=[epoch_metric])
        self._test_history(model, self.metric_names, self.metric_values)

    @skipIf(not torch.cuda.is_available(), "no gpu available")
    def test_batch_metrics_with_multiple_names_returned_by_tensor_on_gpu(self):
        with torch.cuda.device(MetricsModelIntegrationTest.cuda_device):
            batch_metric = get_const_batch_metric(torch.tensor(self.metric_values).cuda())
            model = Model(
                self.pytorch_network,
                self.optimizer,
                self.loss_function,
                batch_metrics=[(self.metric_names, batch_metric)],
            )
            model.cuda()
            self._test_history(model, self.metric_names, self.metric_values)

    @skipIf(not torch.cuda.is_available(), "no gpu available")
    def test_epoch_metrics_with_multiple_names_returned_by_tensor_on_gpu(self):
        with torch.cuda.device(MetricsModelIntegrationTest.cuda_device):
            epoch_metric = ConstEpochMetric(torch.tensor(self.metric_values).cuda())
            model = Model(
                self.pytorch_network,
                self.optimizer,
                self.loss_function,
                epoch_metrics=[(self.metric_names, epoch_metric)],
            )
            model.cuda()
            self._test_history(model, self.metric_names, self.metric_values)

    def test_batch_metrics_with_multiple_names_returned_by_ndarray(self):
        batch_metric = get_const_batch_metric(np.array(self.metric_values))
        model = Model(
            self.pytorch_network, self.optimizer, self.loss_function, batch_metrics=[(self.metric_names, batch_metric)]
        )
        self._test_history(model, self.metric_names, self.metric_values)

    def test_epoch_metrics_with_multiple_names_returned_by_ndarray(self):
        epoch_metric = ConstEpochMetric(np.array(self.metric_values))
        model = Model(
            self.pytorch_network, self.optimizer, self.loss_function, epoch_metrics=[(self.metric_names, epoch_metric)]
        )
        self._test_history(model, self.metric_names, self.metric_values)

    def test_batch_metrics_with_multiple_names_returned_by_list(self):
        batch_metric = get_const_batch_metric(list(self.metric_values))
        model = Model(
            self.pytorch_network, self.optimizer, self.loss_function, batch_metrics=[(self.metric_names, batch_metric)]
        )
        self._test_history(model, self.metric_names, self.metric_values)

    def test_epoch_metrics_with_multiple_names_returned_by_list(self):
        epoch_metric = ConstEpochMetric(list(self.metric_values))
        model = Model(
            self.pytorch_network, self.optimizer, self.loss_function, epoch_metrics=[(self.metric_names, epoch_metric)]
        )
        self._test_history(model, self.metric_names, self.metric_values)

    def test_batch_metrics_with_multiple_names_returned_by_tuple(self):
        batch_metric = get_const_batch_metric(tuple(self.metric_values))
        model = Model(
            self.pytorch_network, self.optimizer, self.loss_function, batch_metrics=[(self.metric_names, batch_metric)]
        )
        self._test_history(model, self.metric_names, self.metric_values)

    def test_epoch_metrics_with_multiple_names_returned_by_tuple(self):
        epoch_metric = ConstEpochMetric(tuple(self.metric_values))
        model = Model(
            self.pytorch_network, self.optimizer, self.loss_function, epoch_metrics=[(self.metric_names, epoch_metric)]
        )
        self._test_history(model, self.metric_names, self.metric_values)

    def test_decomposable_metric_as_epoch_metric(self):
        batch_metric = get_const_batch_metric(tuple(self.metric_values))
        model = Model(
            self.pytorch_network, self.optimizer, self.loss_function, epoch_metrics=[(self.metric_names, batch_metric)]
        )
        self._test_history(model, self.metric_names, self.metric_values)

    def test_predefined_decomposable_metric_as_epoch_metric(self):
        model = Model(self.pytorch_network, self.optimizer, self.loss_function, epoch_metrics=['mse'])
        self._test_history(model, ['mse'], [ANY])

    def test_batch_metrics_with_str_str_tuple(self):
        model = Model(self.pytorch_network, self.optimizer, self.loss_function, batch_metrics=['mse', ('mse2', 'mse')])
        self._test_history(model, ['mse', 'mse2'], [ANY, ANY])

    def test_epoch_metrics_registered_with_unique_name(self):
        names = do_register_metric_class(SomeMetricName, names=['some1', 'some2'], unique_name='unique_some')
        model = Model(
            self.pytorch_network, self.optimizer, self.loss_function, epoch_metrics=['some2', ('something', 'some1')]
        )
        self._test_history(model, ['unique_some', 'something'], [0.0, 0.0])
        unregister_metric_class(names)

    def test_torchmetrics_handling(self):
        expected_names = ['r2_score', 'spearman_corr_coef']
        model = Model(
            self.pytorch_network,
            self.optimizer,
            self.loss_function,
            batch_metrics=[torchmetrics.R2Score(), torchmetrics.SpearmanCorrCoef()],
        )
        self._test_history(model, expected_names, [ANY, ANY])

    def test_repeated_torch_metrics_handling(self):
        expected_names = ['my_const_torch_metric1', 'my_const_torch_metric2']
        model = Model(
            self.pytorch_network,
            self.optimizer,
            self.loss_function,
            batch_metrics=[MyConstTorchMetric(1), MyConstTorchMetric(2)],
        )
        self._test_history(model, expected_names, [1, 2])

    def test_use_torch_metrics_as_epoch_metrics(self):
        expected_names = ['metric1', 'metric2']
        model = Model(
            self.pytorch_network,
            self.optimizer,
            self.loss_function,
            epoch_metrics=[('metric1', MyConstTorchMetric(1)), ('metric2', MyConstTorchMetric(2))],
        )
        self._test_history(model, expected_names, [1, 2])

    @skipIf(not torch.cuda.is_available(), "no gpu available")
    def test_torch_metrics_on_gpu(self):
        with torch.cuda.device(MetricsModelIntegrationTest.cuda_device):
            expected_names = ['r2_score', 'spearman_corr_coef']
            model = Model(
                self.pytorch_network,
                self.optimizer,
                self.loss_function,
                batch_metrics=[torchmetrics.R2Score(), torchmetrics.SpearmanCorrCoef()],
            )
            model.cuda()
            self._test_history(model, expected_names, [ANY, ANY])

    def _test_history(self, model, names, values):
        history = model.fit(
            self.train_x,
            self.train_y,
            validation_data=(self.valid_x, self.valid_y),
            batch_size=MetricsModelIntegrationTest.batch_size,
            epochs=MetricsModelIntegrationTest.epochs,
        )
        for logs in history:
            for name, value in zip(names, values):
                self.assertIn(name, logs)
                self.assertEqual(value, logs[name])
                self.assertIn('val_' + name, logs)
                self.assertEqual(value, logs['val_' + name])

    def test_epoch_metrics_with_str_str_tuple(self):
        dataset_size = MetricsModelIntegrationTest.batch_size * MetricsModelIntegrationTest.steps_per_epoch
        torch.manual_seed(42)
        train_x = torch.rand(dataset_size, 1)
        train_y = torch.randint(10, (dataset_size,))
        valid_x = torch.rand(dataset_size, 1)
        valid_y = torch.randint(10, (dataset_size,))

        model = Model(nn.Linear(1, 10), 'sgd', 'cross_entropy', epoch_metrics=['f1', ('f1_2', 'f1')])
        names = ['fscore_macro', 'f1_2']
        history = model.fit(
            train_x,
            train_y,
            validation_data=(valid_x, valid_y),
            batch_size=MetricsModelIntegrationTest.batch_size,
            epochs=MetricsModelIntegrationTest.epochs,
        )
        for logs in history:
            for name in names:
                self.assertIn(name, logs)
                self.assertEqual(ANY, logs[name])
                self.assertIn('val_' + name, logs)
                self.assertEqual(ANY, logs['val_' + name])

    def test_torch_metrics_with_str_str_tuple(self):
        dataset_size = MetricsModelIntegrationTest.batch_size * MetricsModelIntegrationTest.steps_per_epoch
        torch.manual_seed(42)
        train_x = torch.rand(dataset_size, 1)
        train_y = torch.randint(10, (dataset_size,))
        valid_x = torch.rand(dataset_size, 1)
        valid_y = torch.randint(10, (dataset_size,))

        model = Model(nn.Linear(1, 10), 'sgd', 'cross_entropy', torch_metrics=['f1', ('f1_2', 'f1')])
        names = ['fscore_macro', 'f1_2']
        history = model.fit(
            train_x,
            train_y,
            validation_data=(valid_x, valid_y),
            batch_size=MetricsModelIntegrationTest.batch_size,
            epochs=MetricsModelIntegrationTest.epochs,
        )
        for logs in history:
            for name in names:
                self.assertIn(name, logs)
                self.assertEqual(ANY, logs[name])
                self.assertIn('val_' + name, logs)
                self.assertEqual(ANY, logs['val_' + name])


class MetricsRenamingTest(unittest.TestCase):
    def test_batch_metrics(self):
        actual = rename_doubles(['a', 'a'], [])
        expected = ['a1', 'a2'], []
        self.assertEqual(expected, actual)

        actual = rename_doubles(['a', 'b', 'a', 'a', 'c', 'd'], [])
        expected = ['a1', 'b', 'a2', 'a3', 'c', 'd'], []
        self.assertEqual(expected, actual)

    def test_epoch_metrics(self):
        actual = rename_doubles([], ['a', 'a'])
        expected = [], ['a1', 'a2']
        self.assertEqual(expected, actual)

        actual = rename_doubles([], ['a', 'b', 'a', 'a', 'c', 'd'])
        expected = [], ['a1', 'b', 'a2', 'a3', 'c', 'd']
        self.assertEqual(expected, actual)

    def test_nested_batch_metrics(self):
        actual = rename_doubles([['a', 'a']], [])
        expected = [['a1', 'a2']], []
        self.assertEqual(expected, actual)

        actual = rename_doubles([['a', 'b'], 'b'], [])
        expected = [['a', 'b1'], 'b2'], []
        self.assertEqual(expected, actual)

    def test_nested_epoch_metrics(self):
        actual = rename_doubles([], [['a', 'a']])
        expected = [], [['a1', 'a2']]
        self.assertEqual(expected, actual)

        actual = rename_doubles([], [['a', 'b'], 'b'])
        expected = [], [['a', 'b1'], 'b2']
        self.assertEqual(expected, actual)

    def test_nested_batch_epoch_metrics(self):
        actual = rename_doubles([['a']], [['a']])
        expected = [['a1']], [['a2']]
        self.assertEqual(expected, actual)

        actual = rename_doubles([['a', 'b']], [['c', 'a']])
        expected = [['a1', 'b']], [['c', 'a2']]
        self.assertEqual(expected, actual)

        actual = rename_doubles(['a', 'b', ['b', 'a'], 'c', 'a', 'd'], ['e', ['c', 'a'], 'f', 'a'])
        expected = ['a1', 'b1', ['b2', 'a2'], 'c1', 'a3', 'd'], ['e', ['c2', 'a4'], 'f', 'a5']
        self.assertEqual(expected, actual)
